import logging
import json

from schedules_tools import jsondate
from schedules_tools.models import Task

log = logging.getLogger(__name__)


REPORT_NO_CHANGE = ''
REPORT_ADDED = '_added_'
REPORT_REMOVED = '_removed_'
REPORT_CHANGED = '_changed_'
# REPORT_NEW_VALUE = 'new_value'
# REPORT_OLD_VALUE = 'old_value'

REPORT_KEYS = {
    REPORT_ADDED,
    REPORT_REMOVED,
    REPORT_CHANGED,
}

REPORT_PREFIX_MAP = {
    REPORT_ADDED: '[+]',
    REPORT_REMOVED: '[-]',
    REPORT_CHANGED: '[M]',
    REPORT_NO_CHANGE: 3 * ' ',
}


class ScheduleDiff(object):

    result = []

    subtree_hash_attr_name = 'subtree_hash'

    """ List of attributes used to compare 2 tasks. """
    tasks_match_attrs = ['name', 'dStart', 'dFinish', 'dAcStart', 'dAcFinish', subtree_hash_attr_name]

    def __init__(self, schedule_a, schedule_b):
        self.schedule_a = schedule_a
        self.schedule_b = schedule_b

        self.result = self._diff()

    def __unicode__(self):
        return self.result_to_str()

    def __str__(self):
        return unicode(self).encode('utf-8')

    def result_to_str(self, items=None, level=0, parent_report_type=REPORT_NO_CHANGE):
        """ Textual representation of the diff. """
        res = ''

        if items is None:
            items = self.result

        for item in items:

            if isinstance(item, Task):
                task = item
                child_tasks = task.tasks

                # apply parent_report_type only if NOT REPORT_CHANGED
                if parent_report_type is REPORT_CHANGED:
                    report_type = REPORT_NO_CHANGE
                else:
                    report_type = parent_report_type

            else:
                child_tasks = item['tasks']
                report_type = item['report_type']

                if item['left'] is None or report_type is REPORT_CHANGED:
                    task = item['right']
                else:
                    task = item['left']

            res += '{} {}{}\n'.format(REPORT_PREFIX_MAP[report_type], level * ' ', str(task))

            if child_tasks:
                res += self.result_to_str(child_tasks, level + 1, report_type)

        return res

    def _create_report(self, report_type, left=None, right=None, tasks=[], changed_attrs=[]):
        """
        Returns a dictionary representing a possible change.

            {
                left: Task or None,
                right: Task or None,
                tasks: List of reports from the child Tasks,
                changed_attr: List of changed attributes,
                report_type: Type of change
            }

        """

        if report_type is REPORT_NO_CHANGE and not tasks:
            report = left

        else:
            # No need to keep the whole structure,
            # child tasks will be placed in report['tasks']
            if left is not None:
                left.tasks = []

            if right is not None:
                right.tasks = []

            report = {
                'left': left,
                'right': right,
                'tasks': tasks,
                'changed_attrs': changed_attrs,
                'report_type': report_type,
            }

        return report

    def task_diff(self, task_a, task_b):
        """
        Uses attributes defined in `tasks_match_attrs` to compare 2 tasks and 
        returns a list of atts that don't match.
        """
        return [attr for attr in self.tasks_match_attrs if getattr(task_a, attr) != getattr(task_b, attr)]

    def find_best_match_index(self, task, in_tasks, start_at_index=0):
        """
        Finds the best match for the given `task` in the `in_tasks` and
        returns the index for the best match and a list of the different attributes.
        """
        match_index = None
        unmatched_attrs = self.tasks_match_attrs[:]

        for i in range(start_at_index, len(in_tasks)):
            unmatched = self.task_diff(task, in_tasks[i])

            # NOTE: maybe 'name' should weight more than the other attrs
            if len(unmatched) < len(unmatched_attrs):
                match_index = i
                unmatched_attrs = unmatched

        return match_index, unmatched_attrs

    def _diff(self, tasks_a=None, tasks_b=None):

        if tasks_a is None:
            tasks_a = self.schedule_a.tasks

        if tasks_b is None:
            tasks_b = self.schedule_b.tasks

        res = []
        last_b_index = 0

        # shortcut to create a report for an added task
        def report_task_added(index):
            task = tasks_b[index]
            return self._create_report(REPORT_ADDED, right=task, tasks=task.tasks)

        for i, task in enumerate(tasks_a):
            match_index, diff_attrs = self.find_best_match_index(task, tasks_b, start_at_index=last_b_index)
            report = {}

            if match_index is None:
                report = self._create_report(REPORT_REMOVED, left=tasks_a[i], tasks=tasks_a[i].tasks)

            else:
                # ALL elements between last_b_index and match_index => ADDED
                res.extend([report_task_added(k) for k in range(last_b_index, match_index)])

                # exact match => NO CHANGE
                if len(diff_attrs) == 0:
                    report = self._create_report(REPORT_NO_CHANGE,
                                                 left=task,
                                                 right=tasks_b[match_index])

                # structural change => CHANGED / NO CHANGE
                elif self.subtree_hash_attr_name in diff_attrs:
                    # process child tasks
                    tasks = self._diff(tasks_a[i].tasks, tasks_b[match_index].tasks)

                    report_type = REPORT_CHANGED if len(diff_attrs) > 1 else REPORT_NO_CHANGE
                    report = self._create_report(report_type,
                                                 left=task,
                                                 right=tasks_b[match_index],
                                                 changed_attrs=diff_attrs,
                                                 tasks=tasks)

                # no structural changes => CHANGED
                else:
                    report = self._create_report(REPORT_CHANGED,
                                                 left=task,
                                                 right=tasks_b[match_index],
                                                 changed_attrs=diff_attrs,
                                                 tasks=tasks_b[match_index].tasks)

                last_b_index = match_index + 1

            res.append(report)

        # remaining tasks => ADDED
        res.extend([report_task_added(k) for k in range(last_b_index, len(tasks_b))])

        return res

    def dump_json(self, **kwargs):

        def _encoder(obj):
            if isinstance(obj, Task):
                return obj.dump_as_dict()
            return jsondate._datetime_encoder(obj)

        kwargs['default'] = _encoder
        return json.dumps(self.result, **kwargs)



import datetime
import pytest
import os
import jsondate

from schedules_tools.tests import create_test_schedule
from schedules_tools.converter import ScheduleConverter
from schedules_tools.diff_v2 import ScheduleDiff

SCHEDS_DIRNAME = 'schedule_files'
BASE_DIR = os.path.dirname(os.path.realpath(__file__))
SCHEDS_DIR = os.path.join(BASE_DIR, SCHEDS_DIRNAME)


class TestDiff(object):
    sch1 = None
    sch2 = None

    @pytest.fixture(autouse=True)
    def setUp(self):
        self.sch1 = create_test_schedule()
        self.sch2 = create_test_schedule()

    def test_identical(self):
        diff = self.sch1.diff(self.sch2)

        assert diff == ''

    def test_sch_attrs(self):
        self.sch1.name = '_test_value 1.0.42'
        self.sch1.used_flags = set(['flag1', 'qe'])
        self.sch1.dStart = datetime.datetime(1970, 1, 1)
        self.sch1.dFinish = datetime.datetime(4970, 1, 1)
        self.sch2.changelog['today'] = 'change'

        diff = self.sch1.diff(self.sch2)

        assert diff != ''
        assert 'name' in diff
        assert 'used_flags' in diff
        assert 'dStart' in diff
        assert 'dFinish' in diff
        assert 'changelog' in diff

    def test_task_attrs(self):
        date_delta = datetime.timedelta(1)

        task = self.sch1.tasks[0].tasks[0]
        task.dStart -= date_delta
        task.dFinish += date_delta
        task.link = 'nothing'
        task.note = 'note123'

        diff = self.sch1.diff(self.sch2)

        assert 'link' in diff
        assert 'dStart' in diff
        assert 'dFinish' in diff
        assert 'note' in diff

    def test_missing_task(self):
        diff = self.sch1.diff(self.sch2)
        assert diff == ''

        self.sch2.tasks[0].tasks.pop()

        diff = self.sch1.diff(self.sch2)
        assert diff != ''

    def test_whole_days(self):
        time_delta = datetime.timedelta(minutes=1)
        self.sch2.tasks[0].tasks[1].tasks[0].dFinish -= time_delta

        diff = self.sch1.diff(self.sch2, whole_days=True)
        assert diff == ''


class TestScheduleDiff(object):
    sched_files = ['schedule-diff-a.tjx', 'schedule-diff-b.tjx']

    test_data = [
        pytest.param(('diff_output.json', sched_files), id='json'),
        pytest.param(('diff_output.txt', sched_files), id='text')
    ]

    def import_schedule(self, filename):
        path = os.path.join(SCHEDS_DIR, filename)
        conv = ScheduleConverter()
        return conv.import_schedule(path)

    def diff(self, sched_a_file, sched_b_file):
        sched_a = self.import_schedule(sched_a_file)
        sched_b = self.import_schedule(sched_b_file)
        return ScheduleDiff(sched_a, sched_b)

    @pytest.fixture
    def diff_results(self, request):
        file, scheds_list = request.param
        diff = self.diff(*scheds_list)

        with open(os.path.join(BASE_DIR, 'fixtures', file)) as fd:
            f_ext = os.path.splitext(file)[1]
            if f_ext == '.json':
                yield jsondate.loads(diff.dump_json()), jsondate.load(fd)
            else:
                yield str(diff), fd.read()

    @pytest.mark.parametrize('diff_results', test_data, indirect=True)
    def test_equal_results(self, diff_results):
        actual, expected = diff_results
        assert actual == expected


class TestDiffCLI(object):
    pass



from schedules_tools.schedule_handlers import ScheduleHandlerBase
from schedules_tools import models
import logging
import datetime
import re


COLUMN_TASK_NAME = 'name'
COLUMN_DURATION = 'duration'
COLUMN_START = 'start'
COLUMN_ACSTART = 'acstart'
COLUMN_FINISH = 'finish'
COLUMN_ACFINISH = 'acfinish'
COLUMN_P_COMPLETE = 'complete'
COLUMN_NOTE = 'note'
COLUMN_FLAGS = 'flags'
COLUMN_LINK = 'link'
COLUMN_PRIORITY = 'priority'

log = logging.getLogger(__name__)

smartsheet_logger = logging.getLogger('smartsheet.smartsheet')
smartsheet_logger.setLevel(logging.INFO)

try:
    from smartsheet import Smartsheet
    additional_deps_satistifed = True
except ImportError:
    additional_deps_satistifed = False


class ScheduleHandler_smartsheet(ScheduleHandlerBase):
    provide_export = True
    handle_deps_satisfied = additional_deps_satistifed

    date_format = '%Y-%m-%dT%H:%M:%S'  # 2017-01-20T08:00:00
    _client_instance = None
    _sheet_instance = None
    _re_number = re.compile('[0-9]+')
    columns_mapping_name = {
        'Task Name': COLUMN_TASK_NAME,
        'Duration': COLUMN_DURATION,
        'Start': COLUMN_START,
        'Finish': COLUMN_FINISH,
        '% Complete': COLUMN_P_COMPLETE,
        'Comments': COLUMN_NOTE,
        'Flags': COLUMN_FLAGS,
        'Link': COLUMN_LINK,
        'Actual Start': COLUMN_ACSTART,
        'Actual Finish': COLUMN_ACFINISH,
        'Priority': COLUMN_PRIORITY,
    }

    columns_mapping_id = {}

    def __init__(self, *args, **kwargs):
        super(ScheduleHandler_smartsheet, self).__init__(*args, **kwargs)

    @classmethod
    def is_valid_source(cls, handle=None):
        if not handle:
            handle = cls.handle
        try:
            int(handle)
        except ValueError:
            return False

        return True

    @property
    def client(self):
        if not self._client_instance:
            self._client_instance = Smartsheet(self.options['smartsheet_token'])
            self._client_instance.errors_as_exceptions(True)
        return self._client_instance

    @property
    def sheet(self):
        if not self._sheet_instance:
            self._sheet_instance = self.client.Sheets.get_sheet(self.handle)
        return self._sheet_instance

    def get_handle_mtime(self):
        return self.sheet.modified_at

    def import_schedule(self):
        self.schedule = models.Schedule()
        self.schedule.name = str(self.sheet.name)
        self.schedule.slug = str(self.schedule.unique_id_re.sub('_', self.schedule.name))
        self.schedule.mtime = self.get_handle_mtime()
        parents_stack = []

        # populate columns dict/map
        for column in self.sheet.columns:
            index = self.columns_mapping_name.get(column.title, None)
            if index is None:
                log.debug('Unknown column %s, skipping.' % column.title)
                continue
            self.columns_mapping_id[column.id] = index

        for row in self.sheet.rows:
            self._load_task(row, parents_stack)

        self.schedule.check_top_task()
        self.schedule.dStart = self.schedule.tasks[0].dStart
        self.schedule.dFinish = self.schedule.tasks[0].dFinish
        self.schedule.generate_slugs()
        return self.schedule

    def _parse_date(self, string):
        date = datetime.datetime.strptime(string, self.date_format)
        # We don't require so precise timestamp, so ignore seconds
        date = date.replace(second=0)
        return date

    def _load_task(self, row, parents_stack):
        task = models.Task(self.schedule)
        cells = self._load_task_cells(row)

        task.index = row.row_number
        # task.slug is generated at the end of importing whole schedule

        # skip empty rows
        if not cells[COLUMN_TASK_NAME]:
            return
        task.name = str(cells[COLUMN_TASK_NAME])

        if cells[COLUMN_NOTE]:
            task.note = str(cells[COLUMN_NOTE])
        if COLUMN_PRIORITY in cells:
            task.priority = cells[COLUMN_PRIORITY]

        task.dStart = self._parse_date(cells[COLUMN_START])
        task.dFinish = self._parse_date(cells[COLUMN_FINISH])
        if COLUMN_ACSTART in cells:
            task.dAcStart = cells[COLUMN_ACSTART]
        else:
            task.dAcStart = task.dStart

        if COLUMN_ACFINISH in cells:
            task.dAcFinish = cells[COLUMN_ACFINISH]
        else:
            task.dAcFinish = task.dFinish

        if COLUMN_DURATION in cells:
            # duration cell contains values like '14d', '~0'
            match = re.findall(self._re_number, cells[COLUMN_DURATION])
            if match:
                task.milestone = int(match[0]) == 0
        complete = cells[COLUMN_P_COMPLETE]
        if complete is not None:
            task.p_complete = round(complete * 100, 1)
        if COLUMN_FLAGS in cells and cells[COLUMN_FLAGS]:
            # try first to parse workaround format 'Flags: qe, dev'
            task.parse_extended_attr(cells[COLUMN_FLAGS])

            # then try to consider the value as it is 'qe, dev'
            if not task.flags:
                task.parse_extended_attr(cells[COLUMN_FLAGS],
                                         key=models.ATTR_PREFIX_FLAG)

        if COLUMN_LINK in cells and cells[COLUMN_LINK]:
            # try first to parse workaround format 'Link: http://some.url'
            task.parse_extended_attr(cells[COLUMN_LINK])
            # then try to consider the value as it is 'http://some.url'
            if not task.link:
                task.parse_extended_attr(cells[COLUMN_LINK],
                                         key=models.ATTR_PREFIX_LINK)

        curr_stack_item = {
            'rowid': row.id,
            'task': task
        }
        if not parents_stack:
            # Current task is the topmost (root)
            self.schedule.tasks = [task]
            parents_stack.insert(0, curr_stack_item)
        elif row.parent_id == parents_stack[0]['rowid']:
            # Current task is direct descendant of latest task
            parents_stack[0]['task'].tasks.append(task)
            parents_stack.insert(0, curr_stack_item)
        elif row.parent_id != parents_stack[0]['rowid']:
            # We are currently in the same, or upper, level of tree
            while row.parent_id != parents_stack[0]['rowid']:
                parents_stack.pop(0)
            parents_stack[0]['task'].tasks.append(task)
            parents_stack.insert(0, curr_stack_item)

        task.check_for_phase()
        task.level = len(parents_stack)

    def _load_task_cells(self, row):
        mapped_cells = {}

        for cell in row.cells:
            cell_name = self.columns_mapping_id.get(cell.column_id, None)
            if not cell_name:
                continue

            mapped_cells[cell_name] = cell.value

        return mapped_cells

    def export_schedule(self, output=None):
        sheet_spec = self.client.models.Sheet({
            'name': self.schedule.name,
            'from_id': 5066554783098756,
        })
        resp = self.client.Home.create_sheet_from_template(sheet_spec)
        self.handle = resp.result.id
        column_spec_flags = self.client.models.Column({
            'title': 'Flags',
            'type': 'TEXT_NUMBER',
            'index': 4,
        })
        column_spec_link = self.client.models.Column({
            'title': 'Link',
            'type': 'TEXT_NUMBER',
            'index': 4,
        })
        resp = self.client.Sheets.add_columns(
            self.handle,
            [column_spec_flags, column_spec_link]
        )
        assert resp.message == 'SUCCESS'

        for task in self.schedule.tasks:
            self.export_task(task, parent_id=None)

        # sheet ID
        return self.handle

    @staticmethod
    def calculate_working_days_duration(dstart, dfinish):
        """
        Calculate number of days in starting and ending week.
        https://gist.github.com/archugs/bfbee2b8d210ca07c424#file-workingdaysusingloop-py

        Args:
            dstart: later date
            dfinish: sooner date

        Returns:
            number of working days between dstart and dfinish
        """
        all_days = [dstart + datetime.timedelta(days=x) for x in
                    range((dfinish - dstart).days + 1)]
        working_days = sum(1 for d in all_days if d.weekday() < 5)
        log.debug('duration: {} working days ({} - {})'.format(working_days,
                                                                  dstart,
                                                                  dfinish))
        return working_days

    def export_task(self, task, parent_id=None):
        row = self.client.models.Row()
        row.parent_id = parent_id
        row.to_bottom = True

        if task.name:
            row.cells.append({
                'column_id': self.sheet.columns[0].id,
                'value': task.name
            })
        if task.dStart:
            row.cells.append({
                'column_id': self.sheet.columns[2].id,
                'value': task.dStart.isoformat()
            })
        if task.milestone:
            row.cells.append({
                'column_id': self.sheet.columns[1].id,
                'value': '0'
            })
        elif task.dFinish:
            # finish date can be set only as (start, duration) tuple,
            # put direct value for Finish column is not allowed.
            duration = self.calculate_working_days_duration(task.dStart,
                                                            task.dFinish)
            row.cells.append({
                'column_id': self.sheet.columns[1].id,  # finish #3
                'value': '{}d'.format(duration)
            })
        if task.flags:
            row.cells.append({
                'column_id': self.sheet.columns[4].id,
                'value': ', '.join(task.flags)
            })
        if task.link:
            row.cells.append({
                'column_id': self.sheet.columns[5].id,
                'value': task.link
            })
        if task.p_complete:
            row.cells.append({
                'column_id': self.sheet.columns[8].id,
                'value': task.p_complete / 100.0
            })
        if task.note:
            row.cells.append({
                'column_id': self.sheet.columns[10].id,
                'value': task.note
            })
        resp = self.client.Sheets.add_rows(self.handle, [row])
        assert resp.message == 'SUCCESS', resp.result.message
        assert 1 == len(resp.result)

        row_id = resp.result[0].id

        for nested_task in task.tasks:
            self.export_task(nested_task, parent_id=row_id)

        return row_id
